/*
Copyright (c) 1999-2007, United States Government, as represented by
the Administrator for The National Aeronautics and Space Administration.
All rights reserved.
*/
package gov.nasa.gsfc.drl.rtstps.core.output.rdr;


import gov.nasa.gsfc.drl.rtstps.core.RtStpsException;
import gov.nasa.gsfc.drl.rtstps.core.ccsds.Packet;

import java.util.LinkedList;
import java.util.List;

/**
 * The basic guts of building the RDR HDF infrastructure. This class should be overridden
 * for each specific RDR sensor including the Spacecraft Diary in use.
 * 
 *
 */
public abstract class RDR {
	private RDRName rdrName;
	private RDRAll rdrAll;
	private RDRProduct rdrProduct;
	private LinkedList<RawApplicationPackets> raps = new LinkedList<RawApplicationPackets>();
	private int depth = 20;
	//private int[] appIds = new int[2048];  // 2047 is always idle
	private int setNum = 0; //-1; // UGLY
	private List<Granule> granuleList; // = new LinkedList<Granule>();  // temp possibly instead of re-reading HDF
	public static String DocumentName = "D34862-01_CDFCB-X_Volume_I_PR.pdf"; // default

	private Aggregate aggregate;
	
	private PacketPool packetPool;
	/**
	 * Only used by SpacecraftDiary RDR
	 */
	@Deprecated
	RDR() {
		
	}
	
	/**
	 * Build an RDR around the name, and other items
	 * @param rdrName the RDR name
	 * @param allData the AllData object
	 * @param dataProds the DataProducts object
	 * @param drl the DRL domain
	 * @throws RtStpsException wraps various HDF exceptions
	 */
	public RDR(Stats stats, RDRName rdrName, AllData allData, DataProducts dataProds, FixedDomainDescription drl) throws RtStpsException {
		this.rdrName = rdrName;
		rdrAll = allData.createRDRAll(rdrName);
		
		rdrProduct = dataProds.createRDRProduct(rdrName,
												rdrName.getSensor(), 
												Collection.fromRDRName(rdrName),
												FixedDomainDescription.drl);
		
		packetPool = /*new PacketPoolBySize(rdrName.toString(), stats); */ new PacketPoolHeap(rdrName.toString(), stats);
		
		// make an initial rap...
		
		RawApplicationPackets rap = this.createRawApplicationPackets(rdrName);  //FIXME this is problem for specific sub-classes; their class fields may not be initialized
		
		getRaps().push(rap);
		
		this.setDepth(rdrName.getDepth());
	}

	
	/**
	 * Put a packet into the RDR.
	 * @param p the Packet of interest
	 * @throws RtStpsException wraps any HDF or other exceptions
	 */
	public void put(Packet p) throws RtStpsException {
		
		RawApplicationPackets rap = this.getCurrentRawApplicationPackets();
		
		if (rap.notFull(p)) {
			rap.put(p);
		} else {
			if (rdrAll.write(rap)) {  // this did call rap.close, confusing... 

				// now create the granule because it must refer back to the raps dataspace...
				this.createGranule(rap);

				rap.close();
				getRaps().pop(); // take it off, we are done...
			}

			// start a new rap and put the packet we have in there...
			rap = this.createRawApplicationPackets(rdrName); // need to start a new rap if there are more packet to read...
			rap.notFull(p);
			rap.put(p);			

			if ((getSetNum() % 100) == 0) {
				System.out.println("Passing Creation of Set " + getSetNum());
			}

		}
	}
	
	/**
	 * Override this in the specific specific sensor RawApplicationPacket class
	 * @param rdrName the RDE name
	 * @return the specific RawApplicationPackets for the sensor in the super class
	 */
	protected abstract RawApplicationPackets createRawApplicationPackets(RDRName rdrName);
	
	/**
	 * Get the RDR name
	 * @return the RDRName
	 */
	public RDRName getRDRName() {
		return rdrName;
	}
	
	/**
	 * Get the RDRAll object
	 * @return the RDRAll object
	 */
	public RDRAll getRDRAll() {
		return rdrAll;
	}
	
	/**
	 * Get the RDRProduct
	 * @return the RDRProduct
	 */
	public RDRProduct getRDRProduct() {
		return rdrProduct;
	}
	
	/**
	 * Return a list of RawApplicationsPacket objects that have been created in processing
	 * @return the List of RawApplicationPackets
	 */
	public List<RawApplicationPackets> getAllRawApplicationPackets() {
		return getRaps();
	}

	/**
	 * Get the first (active) RawApplicationsPacket that have been created in processing
	 * @return the first RawApplicationPackets
	 */
	public RawApplicationPackets getCurrentRawApplicationPackets() {
		
		//System.out.println("Raps size: " + raps.size());
		return getRaps().getFirst();
	}
	
	/**
	 * Finish up any local cleanup and close the rdrAll and rdrProduct...
	 * @throws RtStpsException wraps any HDF or other exceptions
	 */
	public void close() throws RtStpsException {
		finish();
		
		// FIXME Don't need this?  each is closed when it is full... I think...
		//for (RawApplicationPackets rap : raps) {
		//	rap.close();
		//}
		// close rdrAll and rdrProducts
		
		packetPool.drain();  // send all the packets back to the heap
		rdrAll.close();
		rdrProduct.close();
	}
	
	/**
	 * Used to create the granule associated with the RawApplicationPacket structure, this method
	 * should be sufficient for most science sensors except for the Spacecraft Diary which is unique
	 * and has its own implementation.
	 * @param rap the RawApplicationPackets associated with a particular sensor
	 * @throws RtStpsException wraps any HDF or other exceptions
	 */
	protected void createGranule(RawApplicationPackets rap) throws RtStpsException {

		// no packets, do nothing...
		if (rap.getTotalPacketCounts() <= 0) return;
		
	

		
		Granule granule = Granule.factory(rap, 
										1L,
										new GranuleId(SpacecraftId.npp, rap.getFirstTime(), 0 /** HACK **/, rap.getTimeSpan()),
										LEOAFlag.Off,
										DocumentName ,
										rap.getPacketTypes(),
										rap.getPacketTypeCounts(),
										new ReferenceId(),
										getSetNum(), 
										rap.getDataSpace(), 
										rdrName);
		//granuleList.add(granule);
		
		rdrProduct.write(granule); // this also calls granule close...

		// this MUST be here or in any specialized implementation
		this.nextSetNum();
		
	}
	
	/**
	 * Build the Aggregate and other wise clean up ...
	 * @throws RtStpsException
	 */
	private void finish() throws RtStpsException {
		RawApplicationPackets rap = this.getCurrentRawApplicationPackets();

		if (rdrAll.write(rap)) {
			this.createGranule(rap);
		} else {
			System.out.println("RDR processing has incomplete granule, skipping...");
		}
		rap.close();
		getRaps().pop(); // remove it, we are done with it
		
		GranuleId beginningGranuleId = rdrProduct.getBeginningGranuleId();
		GranuleId endingGranuleId = rdrProduct.getEndingGranuleId();
		
		// if no granules were processed, just quietly leave and move on...
		if ((beginningGranuleId == null) || (endingGranuleId == null)) return;
		
		aggregate =  Aggregate.factory(beginningGranuleId,
														endingGranuleId,
														rdrProduct.getBeginningOrbit(),
														rdrProduct.getEndingOrbit(),
														rdrProduct.getBeginningDateTime(),
														rdrProduct.getEndingDateTime(),
														rdrProduct.getGranuleCount(),
														rdrName);
		
		rdrProduct.write(aggregate); // FIXME does this close the aggregate out?
	}

	/**
	 * Get the product identifier
	 * @return the ProductIdentifiers
	 */
	public ProductIdentifiers getProductId() {
		return ProductIdentifiers.fromShortName(rdrName.getRDRStringName());
	}

	/**
	 * The list of RawApplicationPackets as LinkedList
	 * @return LinkedList of RawApplicationPackets
	 */
	final LinkedList<RawApplicationPackets> getRaps() {
		return raps;
	}

	/**
	 * A list of the created granule objects
	 * @return List of Granule
	 */
	@Deprecated
	final List<Granule> getGranules() {
		// NOTE: this will return null at at this time
		// if you do not want to return null, create the list for the field above
		// and uncomment the list.add() where the granules are created...
		// it was turned off due to space (heap) issues for some RDRs and is not currently
		// being used by any active classes...
		return granuleList;  
	}
	
	/**
	 * Return the aggregate object that was created if this RDR was closed.
	 * @return Aggregate instance
	 */
	final Aggregate getAggregate() {
		return aggregate;
	}
	/**
	 * Increment the set number
	 */
	final void nextSetNum() {
		this.setNum++;
	}

	/**
	 * Get the current set number
	 * @return the set number in an int
	 */
	final int getSetNum() {
		return setNum;
	}

	/**
	 * The counting depth of RawApplicationPacket, this is associated with an old implementation and probably 
	 * should be deprecated
	 * @param depth the new depth
	 */
	final void setDepth(int depth) {
		this.depth = depth;
	}

	/**
	 * The current depth...
	 * @return the depth in an int
	 */
	final int getDepth() {
		return depth;
	}

	
	final PacketPool getPacketPool() {
		return packetPool;
	}
}
