/*
Copyright (c) 1999-2007, United States Government, as represented by
the Administrator for The National Aeronautics and Space Administration.
All rights reserved.
*/
package gov.nasa.gsfc.drl.rtstps.core.output.rdr;



import gov.nasa.gsfc.drl.rtstps.core.RtStpsException;
import gov.nasa.gsfc.drl.rtstps.core.ccsds.Packet;

import java.io.File;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;



public class RDRFileWriter {
	
	private String tempFilename;
	private String destPath;
	
	// private List<RDRProduct> rdrProducts = new LinkedList<RDRProduct>();
	private RDRCreateFile rdrFile;
	
	
	private int orbit;
	private Origin distributor;
	private MissionName missionName;
	private Origin datasetSource;
	private PlatformShortName platformShortname;
	
	
	public RDRFileWriter(String destPath, 
						int orbit, 
						int rdrCount, 
						Origin distributor, 
						MissionName missionName, 
						Origin datasetSource, 
						PlatformShortName platformShortname) throws RtStpsException {
		this.distributor = distributor;
		this.missionName = missionName;
		this.datasetSource = datasetSource;
		this.platformShortname = platformShortname;
		this.orbit = orbit;
		this.destPath = destPath;
		tempFilename = UUID.randomUUID().toString() + ".h5";
			
		rdrFile = new RDRCreateFile(destPath, tempFilename, rdrCount);
	}
	
	
	public RDRFileWriter(Stats stats, 
			String destPath, 
			int orbit, 
			int rdrCount, 
			Origin distributor, 
			MissionName missionName, 
			Origin datasetSource, 
			PlatformShortName platformShortname) throws RtStpsException {
		
		this.distributor = distributor;
		this.missionName = missionName;
		this.datasetSource = datasetSource;
		this.platformShortname = platformShortname;
		this.orbit = orbit;
		this.destPath = destPath;
		tempFilename = UUID.randomUUID().toString() + ".tmp";

		rdrFile = new RDRCreateFile(stats, destPath, tempFilename, rdrCount);
	}


	public void put(Packet p) throws RtStpsException {
		rdrFile.put(p);
	}


	public void close(boolean hdfCleanup) throws RtStpsException {
		
		Date creationDateAndTime = new Date();
		// the top level attributes
		writeAttributes(creationDateAndTime);
		
		// close the rdr file...
		rdrFile.close();
		
		// clean up any leftover housekeeping
		if (hdfCleanup) {
			HDF5Util.cleanup();
		}
		
		// rename to the final file name
		rename(creationDateAndTime);
	}
	
	private void writeAttributes(Date creationDateAndTime) throws RtStpsException {

		rdrFile.writeAttributes(creationDateAndTime, distributor, this.missionName, this.datasetSource, platformShortname);

	}


	private void rename(Date creationDateAndTime) throws RtStpsException {
		File old = new File(destPath, tempFilename);
		
		List<RDR> rdrs = rdrFile.getRDRs();
		//PDSDate d1 = rdrFile.getFirstDateTime();
		//PDSDate d2 = rdrFile.getLastDataTime();
		
		Aggregate firstAgg = findFirstScienceWeightedAggregate(rdrs);
		
		// LPEATE seems to want the original data based times not the rounded/truncated times in the file name
		PDSDate temp1 = firstAgg.getBeginningDateTime();
		PDSDate d1 = new PDSDate(temp1.getOriginalPacketTime());
		
		//PDSDate temp2 = firstAgg.getEndingDateTime();
		//PDSDate d2 = new PDSDate(temp2.getOriginalPacketTime());
		
		//PDSDate d1 = firstAgg.getBeginningDateTime();
		
		// But this one is offset from it and so is not really a LPEATE or LeapDate
		// it's the same if you get the original time or not in other words.
		// That means we have to substract the leap time from it.
		PDSDate d2 = firstAgg.getEndingDateTime();
		
		
		
		if (rdrs == null) return;
		if (rdrs.size() <= 0) return;
		if ((d1 == null) || (d2 == null)) return;
		
		NPOESSFilename nfn = new NPOESSFilename(rdrs,
												d1,
												d2,
												creationDateAndTime,
												SpacecraftId.npp,
												orbit,
												Origin.nfts);
		
		String realFilename = nfn.toString();
		

		File nnew = new File(destPath, realFilename);
		
		old.renameTo(nnew);
		
		createUserBlock(nnew);
	}
	

	// FIXME only supports either RDRs with "SCIENCE" in the name or the DIARY
	private Aggregate findFirstScienceWeightedAggregate(List<RDR> rdrs) throws RtStpsException {
		List<RDR> sciRDRs = new LinkedList<RDR>();
		RDR diaryRDR = null;
		for (RDR rdr : rdrs) {
			String rdrNameStr = rdr.getRDRName().getRDRStringName();  // avoid toString w/RDRName
			if (rdrNameStr.contains("SCIENCE")) {
				sciRDRs.add(rdr);
			} else if (rdrNameStr.contains("DIARY")) {
				diaryRDR = rdr;
			} else {
				throw new RtStpsException("No support for non-science/non-diary RDRs [" + rdrNameStr + "]");
			}
		}
		
		if ((sciRDRs.size() <= 0) && (diaryRDR == null)) {
			throw new RtStpsException("No science/diary RDRs found -- unable to get first/last time for filename");			
		}
	
		Aggregate firstAggregate = null;
		// if there are any science RDRs, use them...
		if (sciRDRs.size() > 0) {
			// sort by beginningDateTime each aggregate
			// take the earlier one and return that...
			firstAggregate = findFirstAggregate(sciRDRs);
		} else {
			// otherwise must be a non-null diary then right?
			firstAggregate = diaryRDR.getAggregate();
		}
		
		return firstAggregate;
	}

	// loop through and find the earliest aggregate by beginningDateTime
	private Aggregate findFirstAggregate(List<RDR> rdrs) {
		Aggregate firstAgg = null;
		long firstMicros = 0L;
		
		for (RDR rdr : rdrs) {
			Aggregate agg = rdr.getAggregate();
			
			long micros = agg.getBeginningDateTime().getMicrosSinceEpoch();
			
			if (firstMicros == 0L) {
				firstMicros = micros;
				firstAgg = agg;
			} else if (micros < firstMicros) {
				firstMicros = micros;
				firstAgg = agg;
			}
		}
		return firstAgg;
	}


	private void createUserBlock(File file) throws RtStpsException {
		
		List<RDR> rdrs = rdrFile.getRDRs();
		
		if (rdrs == null) return;
		
		if (rdrs.size() <= 0) return;
		
		UserBlock userBlock = new UserBlock(MissionName.NPP, PlatformShortName.NPP, rdrs);
		
		userBlock.write(file);
		
		userBlock.close();
	}

	public String toString() {
		return NPOESSFilename.productIdsToString(rdrFile.getRDRs());
	}
}
