/*
Copyright (c) 1999-2007, United States Government, as represented by
the Administrator for The National Aeronautics and Space Administration.
All rights reserved.
*/
package gov.nasa.gsfc.drl.rtstps.core.output.rdr;

import gov.nasa.gsfc.drl.rtstps.core.RtStpsException;
import gov.nasa.gsfc.drl.rtstps.core.ccsds.Packet;

import java.util.EnumSet;


/**
 * Implements an algorithm for collecting VIIRS segmented or group packets into a cohesive unit so that blocks of them
 * maybe written atomically into the RawApplicationPackets area for a VIIRS RDR file.  The goal then of this class is to
 * keep the segments together in each RawApplicationPackets and to put several together as designated before contents are
 * then ready to be written to the RDR file. 
 * 
 * This makes multiple 48 scan VIIRS granules in one file
 */
public class VIIRSRawApplicationPackets_MultiGranule extends RawApplicationPackets {
	private long currentScanNumber = -1;
	private int scanCounter = 0;
	private int scansPerGranule;
	private EnumSet<PacketName> sciPacketRange = EnumSet.range(PacketName.M04, PacketName.DNB_LGS); //FIXSET1
	private EnumSet<PacketName> engPacketRange = EnumSet.range(PacketName.ENG, PacketName.ENG);
	private EnumSet<PacketName> calPacketRange = EnumSet.range(PacketName.CAL, PacketName.CAL);
	private Stats stats;
	private int sciPacketCount = 0;
	private long lastCalPacketScanNumber = -999;
	private long lastEngPacketScanNumber = -999;
	private int lastEngPacketIndex;
	private int lastCalPacketIndex;
	
	/**
	 * Constructor for creating an nth instance of a VIIRS raw application data packet area
	 * @param satellite  the name of the spacecraft
	 * @param setNum  the set number
	 */
	public VIIRSRawApplicationPackets_MultiGranule(SpacecraftId satellite, int setNum, int scansPerGranule, PacketPool packetPool) {
		super(satellite, RDRName.VIIRS_Science, setNum, packetPool);
		
		if (setNum < 0) {
			throw new IllegalArgumentException("Illegal Index [" + setNum + "]");
		}
		//System.out.println("Constructor: Scans per granule: " + this.scansPerGranule);
		this.scansPerGranule = scansPerGranule;
	}
	
	public VIIRSRawApplicationPackets_MultiGranule(Stats stats, SpacecraftId satellite, int setNum, int scansPerGranule, PacketPool packetPool) {
		super(satellite, RDRName.VIIRS_Science, setNum, packetPool);
		
		if (setNum < 0) {
			throw new IllegalArgumentException("Illegal Index [" + setNum + "]");
		}
		//System.out.println("Constructor: Scans per granule: " + this.scansPerGranule);
		this.scansPerGranule = scansPerGranule;
		
		this.stats = stats;
	}
	
	/**
	 * Constructor which attempts to read the RawApplicationPacket entry that pre-exists.
	 * The contents of the dataspace are read into a memory buffer... assuming it will fit.
	 * @param allRDRId  the rdrAll Groups id
	 * @param setNum the set number of raw entry 
	 * @throws RtStpsException 
	 */
	public VIIRSRawApplicationPackets_MultiGranule(int allRDRId, int setNum) throws RtStpsException {
		super(allRDRId, setNum);
	}
	
	public VIIRSRawApplicationPackets_MultiGranule(int readId, int setNum, boolean usedByGranuleOnly) throws RtStpsException {
		super(readId, setNum, true);
	}



	/**
	 * Determine if the object will take more packets.  For VIIRS there
	 * are two cases.  First if the collection of packets into this particular
	 * RawApplication has not started, its considered to be not-full.
	 * Second if the scan number matches the current scan number, it is considered
	 * not-full.   If the collection has started and the scan number changes, it 
	 * is considered full and any collected packets should be written to the HDF
	 * @param p packet to be written
	 * @return true or false
	 * @throws RtStpsException 
	 */
	public boolean notFull(Packet p) throws RtStpsException {
		RDRName rdrName = RDRName.fromAppId(p.getApplicationId());
		
		if (rdrName != RDRName.VIIRS_Science) {
			throw new RtStpsException("Not a VIIRS packet [" + p.getApplicationId() + "]");
		}
		
		// A collection is defined for our purposes as all the packets associated with a particular scan number
		// for VIIRS.
		// However at the start of the packet stream, this may not be clear and so there is the special
		// case of having no scan number yet but a valid VIIRs packet.  This will be collected and result in
		// a 'runt' VIIRS granule...
		//
		
		//System.out.println("Scans per granule: " + this.scansPerGranule);
		
		// to start: assume it is  not full...
		boolean notFullStatus = true;
		boolean scanCountTransition = false;  // from scan counter field value 1 to 2 say...
		
		// look for scan number transitions to signal the start and end of the collections
		// a special case occurs for the first packets read as it is likely this set does not
		// start at a particular scan boundary and so these packets are collected
		long scanNumber = -1L;
		if (p.isFirstPacketInSequence() || p.isStandalonePacket()) {
			
			// the scan number changes only in the science packets, the other calibration/eng
			// are associated with the current scan...
			//System.out.print(p.getApplicationId()+" ");
			if (isVIIRSSciencePacket(p)) {
				scanNumber = VIIRSSciencePacket.getScanNumber(p);
			}
			if (isVIIRSEnggPacket(p)) {
				
					
				scanNumber = ViirsEnggPacket.getScanNumber(p);
				//is this the first engineering packet we are receiving
				if(lastEngPacketScanNumber==-999){
					//Check if we have already received a Calibration Packet
					if(lastCalPacketScanNumber>=0){
						//We will have to remove all packets with this scan number
						System.out.println("******Need to remove packets with scan-number "+lastCalPacketScanNumber+" scanCounter="+scanCounter);
						packetPool.flush(getPacketList());
						scanCounter=0;
						setFirstTime(0L);
					}
					lastEngPacketScanNumber=scanNumber;	
				}
			}
			if (isVIIRSCalPacket(p)) {
				scanNumber = VIIRSCalibrationPacket.getScanNumber(p);
			}
			//System.out.print("ScanNumber="+scanNumber+" | ");
			
			if (scanNumber != currentScanNumber) {
				//System.out.print(p.getApplicationId()+" ");
				//System.out.print("ScanNumber="+scanNumber+" | ");
					scanCountTransition = true;
			}
			
			
		}
		
		// if the packet signals a scan transition
		if (scanCountTransition) {
			++scanCounter;  // increment the running scan count
			//System.out.print("scanCounter: "+ scanCounter+" | ");
			// if its the FIRST one, that is we are going from -1 currentScanNumber to the first scan
			/// if it is the first one and the packet list > 0 in size,  then "runt" packets were stored first
			/// in theory we could throw them all out but won't now.  We just assume it is full and the
			/// caller will store a stubby first granule
			
			// else if it isn't ...
			/// then it's "normal" and we check if the scan group is full
			
			if (currentScanNumber < 0) {
				// special case -- the first few packets
				// Lock may have occurred mid-scan say, so those fell through these
				// checks and were simply piled up on the list.
				// Now a new scan scan number has finally arrived, so that partial scan
				// will be put into the first granule as a 'stubby'...
				if (getPacketList().size() > 0) {
					// so if there are packets and this is the first new scan number
					// it marked as full.  This will create (assuming that write is then called)
					// the granule_0 as a "short" or "stubby" granule without the full "scanCounts"
					// in it
					
					//System.out.println("First scan found in VIIRS -- runt packets proceeds it: " + scanCounter);
					
					
					//getPacketList().clear(); // TEST -- wipe out the partial scan & keep going... ?
					packetPool.flush(getPacketList());  // don't wipe it out, drain it back to the pool
					
					//currentScanNumber = scanNumber;  // ORIGINAL
					scanCounter = 0; // TEST
					
					//notFullStatus = false; // this is the one that matters, this will return full and then the write should follow...
					notFullStatus = true;
				} else {
					//System.out.println("First scan found in VIIRS -- no packet proceeds it:" + scanCounter);
					currentScanNumber = scanNumber;
					//scanCounter = 0;
					notFullStatus = true;
				}
			} else if (scanCounter > scansPerGranule) {
				// If the scansPerGranule have been reached
				// the collected packets should be written first
				// before starting again
				//System.out.println("Scan transition VIIRS -- group complete. [" + scanCounter + "]");
				--scanCounter;  // subtract off the one we aren't actually collecting for since it's full
				notFullStatus = false;
				//System.out.println("Granule full**");
			} else {
				//if (scanCounter == 1) {
				//	System.out.println("Scan transition VIIRS -- new scan group. [" + scanCounter + "]");
				//} else {
				//	System.out.println("Scan transition VIIRS -- scan number remains in place -- counter: [" + scanCounter + "]");
				//}
				
				currentScanNumber = scanNumber;  
				// not full status is true here...
				notFullStatus = true;

			}
			
		} else {
			// if it does not it is stored...
			notFullStatus = true;
		}
		
		
	
		return notFullStatus;
	}
	
	/**
	 * Determine if the object will take more packets.  For VIIRS there
	 * are two cases.  First if the collection of packets into this particular
	 * RawApplication has not started, its considered to be not-full.
	 * Second if the scan number matches the current scan number, it is considered
	 * not-full.   If the collection has started and the scan number changes, it 
	 * is considered full and any collected packets should be written to the HDF
	 * @param p packet to be written
	 * @return true or false
	 * @throws RtStpsException 
	 */
	public boolean notFull_OLD(Packet p) throws RtStpsException {
		RDRName rdrName = RDRName.fromAppId(p.getApplicationId());
		
		if (rdrName != RDRName.VIIRS_Science) {
			throw new RtStpsException("Not a VIIRS packet [" + p.getApplicationId() + "]");
		}
		
		// A collection is defined for our purposes as all the packets associated with a particular scan number
		// for VIIRS.
		// However at the start of the packet stream, this may not be clear and so there is the special
		// case of having no scan number yet but a valid VIIRs packet.  This will be collected and result in
		// a 'runt' VIIRS granule...
		//
		
		
		// to start: assume it is  not full...
		boolean notFullStatus = true;
		
		// look for scan number transitions to signal the start and end of the collections
		// a special case occurs for the first packets read as it is likely this set does not
		// start at a particular scan boundary and so these packets are collected
		
		if (p.isFirstPacketInSequence() || p.isStandalonePacket()) {
			
			// the scan number changes only in the science packets, the other calibration/eng
			// are associated with the current scan...
			long scanNumber = currentScanNumber;
			if (isVIIRSSciencePacket(p)) {
				scanNumber = VIIRSSciencePacket.getScanNumber(p);
			}
			if (currentScanNumber < 0) {
				// special case -- the first few packets
				// Lock may have occurred mid-scan say, so those fell through these
				// checks and were simply piled up on the list.
				// Now a new scan scan number has finally arrived, so that partial scan
				// will be put into the first granule as a 'stubby'...
				if (getPacketList().size() > 0) {
					// so if there are packets and this is the first new scan number
					// it marked as full.  This will create (assuming that write is then called)
					// the granule_0 as a "short" or "stubby" granule without the full "scanCounts"
					// in it
					
					// TEST -- wipe out the partial scan & keep going... ?
					//getPacketList().clear();
					//System.out.println("I will throw out early VIIRS because it is not a full granule.");
					// ORIGINAL
					currentScanNumber = scanNumber;  
					scanCounter = 0;
					//notFullStatus = false; // ORIGINAL this is the one that matters, this will return full and then the write should follow...
					notFullStatus = true;  // if the first standalone pkt occurs but the list had stuff on it, we end up here... wiping the list, so it would not be full
				} else {
					currentScanNumber = scanNumber;
					scanCounter = 0;
					notFullStatus = true;
				}
			} else if (scanNumber != currentScanNumber ) {
				// transition from old scan number, then increment scanCounter
				
				// If the scansPerGranule have been reached
				// the collected packets should be written first
				// before starting again
				if (++scanCounter > scansPerGranule) {
					notFullStatus = false;
				} else {
					currentScanNumber = scanNumber;
					// not full status is true here...
					notFullStatus = true;
				}
			}
		} // else {
			// it is accepted without question if it is not a standalone or first packet or part of a partial first scan, notFullStatus is true
		// }
		
	
		return notFullStatus;
	}
	
	// uses the PacketName set of VIIRS packet enums to check that the packet given is in range
	// OR we could just hardcode the APIDs here but... this so much better obviously right? :-)
	private boolean isVIIRSSciencePacket(Packet p) {
		return sciPacketRange.contains(PacketName.fromAppId(p.getApplicationId()));
	}
	private boolean isVIIRSEnggPacket(Packet p) {
		return engPacketRange.contains(PacketName.fromAppId(p.getApplicationId()));
	}
	private boolean isVIIRSCalPacket(Packet p) {
		return calPacketRange.contains(PacketName.fromAppId(p.getApplicationId()));
	}
	/**
	 * The packet is stored until a packet transition which is checked in the notFull() method.
	 * When full the RawApplicationPacket should then be written to the HDF
	 * @param p VIIRS science packet to be written
	 * @throws RtStpsException the packet is not a VIIRS science packet or the scan number does not match previous packets (notFull was not called first)
	 */
	public void put(Packet p) throws RtStpsException {
		
		sciPacketCount++;
		
		this.updateAppIdCounters(p.getApplicationId());
		
		RDRName rdrName = RDRName.fromAppId(p.getApplicationId());
		
		if (rdrName != RDRName.VIIRS_Science) {
			throw new RtStpsException("Not a VIIRS science packet [" + p.getApplicationId() + "]");
		}
	
		if ((currentScanNumber < 0) && (isVIIRSSciencePacket(p) == false)) {
			// no scan lock yet, this is could be APID=826 or maybe 825 from a the previous scan
			// so we throw it out...
			return;
		}
		
		if (p.hasSecondaryHeader()) {
			if (currentScanNumber < 0) {
				// this handles the case that the collection was full, and
				// the block of packets was written...  and then the packet
				// that was trying to be written is written immediately after this
				// 
				currentScanNumber = VIIRSSciencePacket.getScanNumber(p);
			}
			// secondary time stamp only valid in in first or standalone packets...
			if (getFirstTime() == 0L) {
				//System.out.println("Here...");
				setFirstTime(p.getTimeStamp(8));
			}
			
			setLastTime(p.getTimeStamp(8));
			
			if (getFirstSystemTime() == 0L) {
				
				setFirstSystemTime(PDSDate.getCurrentSystemTimeMicrosSinceEpoch());
			}
			
			setLastSystemTime(PDSDate.getCurrentSystemTimeMicrosSinceEpoch());

		}
		// the packet is deep copied, it is allocated off a pool
		// which may hold previously copied packets...  they must be put
		// back on the pool when done...
		
		Packet pcopy = CopyPacket.deep(p, packetPool);
		
		getPacketList().add(pcopy);
		
		//Was the packet put into list a Calibration or Engg Packet
		// If so record their scan no.
		if (p.isFirstPacketInSequence() || p.isStandalonePacket())
		{
			if (isVIIRSEnggPacket(p)) {
				lastEngPacketScanNumber = ViirsEnggPacket.getScanNumber(p);
				lastEngPacketIndex = getPacketList().size()-1;
			}
			if (isVIIRSCalPacket(p)) {
				lastCalPacketScanNumber = VIIRSCalibrationPacket.getScanNumber(p);
				lastCalPacketIndex = getPacketList().size()-1;
			}
		}
		
	}
	
	
	/**
	 * Write the collected group of packets to the designated HDF file using the handle
	 * @return true if the RawApplicationPacket was written, false if not
	 * @exception any HDF exceptions are wrapped in an RtStpsException
	 */
	public boolean write(int hdfFile) throws RtStpsException {
		boolean ret = true;
		
		//System.out.println("In Write -- " + this.packetPool.toString());
		
		if (scanCounter == scansPerGranule) {  // we want 48 scans exactly
			
			// write returns true/false if it actually wrote anything
			// its ignored here as the scanCounter would not be at scansPerGranule
			// if there no packets...
			// Maybe it should it be double checked just to make sure?  
			System.out.println("******LastCalPacketScanNo="+lastCalPacketScanNumber+" LastEngPacketScanNumber="+lastEngPacketScanNumber);
			super.write(hdfFile); 
		
			if (stats == null) {
				System.out.println("Creating VIIRS Granule -- [" + this.scanCounter + "] scans");
			} else {
				stats.sci_createdGranules.value++;
			}
			
			ret = true;
		} else if (scanCounter < scansPerGranule){
			System.out.println("*****LastCalPacketScanNo="+lastCalPacketScanNumber+" LastEngPacketScanNumber="+lastEngPacketScanNumber);
			if(lastEngPacketScanNumber>lastCalPacketScanNumber)
			packetPool.removeEnd(getPacketList(),lastEngPacketIndex);
			if(getPacketList().size()>0){// make sure that wasn't the only packet
			    super.write(hdfFile); 
			}
			
			if (stats == null) {
				System.out.println("Creating partial VIIRS Granule -- [" + this.scanCounter + "] scans");
			} else {
				stats.sci_createdGranules.value++;
			}
			
			
			ret = true;
			
			/*** 
			if (stats == null) {
				System.out.println("Deleting partial VIIRS Granule -- [" + this.scanCounter + "] scans");
			} else {
				stats.sci_discardedGranules.value++;
			}

			ret = false;
			***/
			
			// Note it and write it any way
		} else if (scanCounter > scansPerGranule) {
			if (stats == null) {
				System.out.println("Too many scans in the VIIRS Granule, something is wrong -- [" + this.scanCounter + "] scans");
			} else {
				stats.sci_discardedGranules.value++;
			}

			ret = false;
		}
		
		System.out.println(">> VIIRS Sci packets in granule: " + sciPacketCount + " Packet per scan: " + (float)sciPacketCount / (float)scansPerGranule);
		System.out.println(">> VIIRS app ids = " + super.appIdCountersToString(scansPerGranule));
		sciPacketCount = 0;
		this.currentScanNumber = -1;
		this.scanCounter = 0;
		

		return ret;
	}
	
	
	/**
	 * Close out the RawApplicationPacket which writes the results to the HDF file and cleans up.
	 * @exception any HDF exceptions are wrapped in an RtStpsException
	 */
	public void close() throws RtStpsException {
		super.close();
		this.currentScanNumber = -1;  // no point of this..
		this.scanCounter = 0; // nor this...
		sciPacketCount = 0;
		//System.out.println("End of Close -- " + this.packetPool.toString());

	}

}
